/* Copyright (c) 2009 InMon Corp. ALL RIGHTS RESERVED */
/* License: http://www.inmon.com/products/virtual-probe/license.php */

#if defined(__cplusplus)
extern "C" {
#endif

#include "vsprobed.h"

#define VSP_MAX_LINELEN 2048
#define VSP_MAX_CONFIG_DEPTH 3
#define VSP_SEPARATORS " \t\r\n=;"
  
  typedef enum { VSPTOKENTYPE_UNDEFINED=0,
		 VSPTOKENTYPE_SYNTAX,
		 VSPTOKENTYPE_OBJ,
		 VSPTOKENTYPE_ATTRIB } EnumVSPTokenType;
  
  // read the special tokens include twice,
  // first to generate the enum, then to
  // populate the special token lookup table
  typedef enum {
#define VSPTOKEN_DATA(tag, token, type) tag,
#include "vsprobetokens.h"
#undef VSPTOKEN_DATA
    VSPTOKEN_NUM_TOKENS } EnumVSPSpecialToken;
  
  typedef struct _VSPSpecialToken {
    EnumVSPSpecialToken tag;
    char *str;
    EnumVSPTokenType type;
  } VSPSpecialToken;
  
  static const VSPSpecialToken VSPSpecialTokens[] = {
#define VSPTOKEN_DATA(tag, token, type) { tag, token, type },
#include "vsprobetokens.h"
#undef VSPTOKEN_DATA
  };


  typedef struct _VSPToken {
    struct _VSPToken *nxt;
    u_int32_t lineNo;
    char *str;
    EnumVSPSpecialToken stok;
  } VSPToken;
  
  typedef enum { VSPOBJ_VSP=0,
		 VSPOBJ_SFLOW,
		 VSPOBJ_COLLECTOR,
		 VSPOBJ_SWITCH,
		 VSPOBJ_PCAP,
		 VSPOBJ_INTERFACE } EnumVSPObject;


#define ADD_TO_LIST(linkedlist, obj) \
  do { \
    obj->nxt = linkedlist; \
    linkedlist = obj; \
  } while(0)
    
  /*________________---------------------------__________________
    ________________       lookupAddress       __________________
    ----------------___________________________------------------
  */

  int lookupAddress(char *name, struct sockaddr *sa, SFLAddress *addr, int family)
  {
    struct addrinfo *info = NULL;
    struct addrinfo hints;
    memset(&hints, 0, sizeof(hints));
    hints.ai_socktype = SOCK_DGRAM; // constrain this so we don't get lots of answers
    hints.ai_family = family; // PF_INET, PF_INET6 or 0
    int err = getaddrinfo(name, NULL, &hints, &info);
    if(err) {
      switch(err) {
      case EAI_NONAME: break;
      case EAI_NODATA: break;
      case EAI_AGAIN: break; // loop and try again?
      default: syslog(LOG_ERR, "getaddrinfo() error: %s\n", gai_strerror(err)); break;
      }
      return NO;
    }
  
    if(info == NULL) return NO;
  
    if(info->ai_addr) {
      // answer is now in info - a linked list of answers with sockaddr values.
      // extract the address we want from the first one.
      switch(info->ai_family) {
      case PF_INET:
	{
	  struct sockaddr_in *ipsoc = (struct sockaddr_in *)info->ai_addr;
	  addr->type = SFLADDRESSTYPE_IP_V4;
	  addr->address.ip_v4 = ipsoc->sin_addr;
	  if(sa) memcpy(sa, info->ai_addr, info->ai_addrlen);
	}
	break;
      case PF_INET6:
	{
	  struct sockaddr_in6 *ip6soc = (struct sockaddr_in6 *)info->ai_addr;
	  addr->type = SFLADDRESSTYPE_IP_V6;
	  addr->address.ip_v6 = ip6soc->sin6_addr;
	  if(sa) memcpy(sa, info->ai_addr, info->ai_addrlen);
	}
	break;
      default:
	syslog(LOG_ERR, "get addrinfo: unexpected address family: %d\n", info->ai_family);
	return NO;
	break;
      }
    }
    // free the dynamically allocated data before returning
    freeaddrinfo(info);
    return YES;
  }

  /*________________---------------------------__________________
    ________________       parseMAC            __________________
    ----------------___________________________------------------
  */

  static u_char hex2bin(u_char c)
  {
    return (isdigit(c) ? (c)-'0': ((toupper(c))-'A')+10)  & 0xf;
  }
  
  
  int hexToBinary(u_char *hex, u_char *bin, u_int32_t binLen)
  {
    // read from hex into bin, up to max binLen chars, return number written
    u_char *h = hex;
    u_char *b = bin;
    u_char c;
    u_int32_t i = 0;
    
    while((c = *h++) != '\0') {
      if(isxdigit(c)) {
	u_char val = hex2bin(c);
	if(isxdigit(*h)) {
	  c = *h++;
	  val = (val << 4) | hex2bin(c);
	}
	*b++ = val;
	if(++i >= binLen) return i;
      }
      else if(c != '.' &&
	      c != '-' &&
	      c != ':') { // allow a variety of byte-separators
	return i;
      }
    }
    return i;
  }

  int parseMAC(char *str, u_int64_t *mac)
  {
    u_char macbytes[6];
    if(hexToBinary((u_char *)str, macbytes, 6) != 6) return NO;
    // cast to 64-bit integer by simply copying in the bytes.
    // It doesn't matter whether the architecture is big endian
    // or little endian, we are just using this as a convenient
    // comparison symbol.
    memcpy(mac, macbytes, 6);
    return YES;
  }

  /*_________________---------------------------__________________
    _________________      parseError           __________________
    -----------------___________________________------------------
  */

  static void parseError(VSP *sp, VSPToken *tok, char *msg1, char *msg2)
  {
    syslog(LOG_ERR, "parse error at <%s><%s> on line %d of %s : %s %s",
	   tok->str,
	   tok->nxt ? tok->nxt->str : "",
	   tok->lineNo,
	   sp->configFile,
	   msg1,
	   msg2);
  }


  /*_________________---------------------------__________________
    _________________   attribute extraction    __________________
    -----------------___________________________------------------
  */

  // expectToken

  static VSPToken *expectToken(VSP *sp, VSPToken *tok, EnumVSPSpecialToken stok)
  {
    VSPToken *t = tok;
    t = t->nxt;
    if(t == NULL || t->stok != stok) {
      parseError(sp, tok, "expected ", VSPSpecialTokens[stok].str);
      return NULL;
    }
    return t;
  }

  // expectInteger32

  static u_int32_t getMultiplier(char *str)
  {
    u_int32_t mult = 1;
    u_int32_t len = strlen(str);
    char last = toupper(str[len - 1]);
    if(last == 'K' || last == 'M' || last == 'G') {
      // number of the form "100M" or "1G"
      str[len - 1] = '\0'; // blat the K, M or G
      if(last == 'K') mult = 1000;
      if(last == 'M') mult = 1000000;
      if(last == 'G') mult = 1000000000;
    }
    return mult;
  }

  static VSPToken *expectInteger32(VSP *sp, VSPToken *tok, u_int32_t *arg, u_int32_t minVal, u_int32_t maxVal)
  {
    VSPToken *t = tok;
    t = t->nxt;
    if(t == NULL || !isdigit(t->str[0])) {
      parseError(sp, tok, "expected integer", "");
      return NULL;
    }
    char *str = strdup(t->str); // take a copy so we can modify it
    u_int32_t mult = getMultiplier(str);
    *arg = (mult * strtol(str, NULL, 0));
    free(str);
    if(*arg < minVal || *arg > maxVal) {
      parseError(sp, tok, "range error", "");
      return NULL;
    }
    return t;
  }

  // expectInteger64

  static VSPToken *expectInteger64(VSP *sp, VSPToken *tok, u_int64_t *arg)
  {
    VSPToken *t = tok;
    t = t->nxt;
    if(t == NULL || !isdigit(t->str[0])) {
      parseError(sp, tok, "expected integer", "");
      return NULL;
    }
    char *str = strdup(t->str); // take a copy so we can modify it
    u_int32_t mult = getMultiplier(str);
    *arg = ((u_int64_t)mult * (u_int64_t)strtol(str, NULL, 0));
    free(str);
    return t;
  }

  // expectIP

  static VSPToken *expectIP(VSP *sp, VSPToken *tok, SFLAddress *addr, struct sockaddr *sa)
  {
    VSPToken *t = tok;
    t = t->nxt;
    if(t == NULL || lookupAddress(t->str, sa, addr, 0) == NO) {
      parseError(sp, tok, "expected hostname or IP", "");
      return NULL;
    }
    return t;
  }

  // expectMAC

  static VSPToken *expectMAC(VSP *sp, VSPToken *tok, u_int64_t *macPtr)
  {
    VSPToken *t = tok;
    t = t->nxt;
    if(t == NULL ||  parseMAC(t->str, macPtr) == NO) {
      parseError(sp, tok, "expected MAC address", "");
      return NULL;
    }
    return t;
  }

  // expectStr

  static VSPToken *expectStr(VSP *sp, VSPToken *tok, char *paramName, char **param)
  {
    VSPToken *t = tok;
    t = t->nxt;
    if(t == NULL || !isalnum(t->str[0])) {
      parseError(sp, tok, "expected", paramName);
      return NULL;
    }
    *param = t->str;
    return t;
  }

  /*_________________---------------------------__________________
    _________________     new object fns        __________________
    -----------------___________________________------------------
  */

  static VSPSFlowSettings *newSFlowSettings(VSPSFlow *sf) {
    VSPSFlowSettings *st = (VSPSFlowSettings *)calloc(1, sizeof(VSPSFlowSettings));
    st->samplingRate = SFL_DEFAULT_SAMPLING_RATE;
    st->pollingInterval = SFL_DEFAULT_POLLING_INTERVAL;
    return st;
  }

  static VSPSFlow *newSFlow(VSP *sp) {
    VSPSFlow *sf = (VSPSFlow *)calloc(1, sizeof(VSPSFlow));
    sf->sFlowSettings = newSFlowSettings(sf);
    sf->firstSubAgentId = 0;
    sf->firstIfIndex = 1;
    // park the 64-bit integer representing the broadcast address here
    parseMAC("FF:FF:FF:FF:FF:FF", &sf->macBroadcast);
    sp->sFlow = sf; // just one of these, not a list
    sf->myVSP = sp;
    return sf;
  }

  static VSPCollector *newCollector(VSPSFlow *sf) {
    VSPCollector *col = (VSPCollector *)calloc(1, sizeof(VSPCollector));
    ADD_TO_LIST(sf->collectors, col);
    sf->numCollectors++;
    col->udpPort = SFL_DEFAULT_COLLECTOR_PORT;
    return col;
  }

  static VSPSwitch *newSwitch(VSPSFlow *sf) {
    VSPSwitch *sw = (VSPSwitch *)calloc(1, sizeof(VSPSwitch));
    ADD_TO_LIST(sf->switches, sw);
    sw->mySFlow = sf;
    sw->subAgentId = sf->firstSubAgentId + sf->numSwitches++;
    return sw;
  }

  static VSPInterface *newInterface(VSPSwitch *sw) {
    VSPInterface *intf = (VSPInterface *)calloc(1, sizeof(VSPInterface));
    ADD_TO_LIST(sw->interfaces, intf);
    intf->ifIndex = sw->mySFlow->firstIfIndex + sw->numInterfaces++;
    intf->ifType = 6; // ethernetcsmacd
    intf->ifSpeed = 0;
    intf->mySwitch = sw;
    return intf;
  }

  static VSPPCAP *newPCAP(VSPSwitch *sw) {
    sw->pktReader = (VSPPCAP *)calloc(1, sizeof(VSPPCAP));
    // set defaults
    sw->pktReader->promiscuous = VSP_DEFAULT_PROMISCUOUS;
    sw->pktReader->snaplen = SFL_DEFAULT_HEADER_SIZE;
    sw->pktReader->batch = VSP_DEFAULT_READ_BATCH;
    return sw->pktReader;
  }

  static VSPToken *newToken(char *str, int len) {
    VSPToken *token = (VSPToken *)calloc(1, sizeof(VSPToken));
    token->str = (char *)calloc(1, len + 1);
    memcpy(token->str, str, len);
    // if it is special token, then record the index number here too
    for(u_int32_t i = 0; i < VSPTOKEN_NUM_TOKENS; i++) {
      if(strcasecmp(token->str, VSPSpecialTokens[i].str) == 0) {
	token->stok = (EnumVSPSpecialToken)i;
      }
    }
    return token;
  }

  /*_________________---------------------------__________________
    _________________      readTokens           __________________
    -----------------___________________________------------------
  */

  static VSPToken *nextToken(char *p, char **out)
  {
    char *r = p;
    VSPToken *token = NULL;

    // skip separators
    r += strspn(r, VSP_SEPARATORS);

    if(*r != '\0') {
      // found token, but watch out for a contiguous '{' or '}' token.
      u_int32_t len = strcspn(r, "{}" VSP_SEPARATORS);
      if(len == 0) len = 1; // started with '{' or '}'
      token = newToken(r, len);
      r += len;
    }
    // tell the caller how far we got
    *out = r;
    // return token or NULL
    return token;
  }

  static  VSPToken *reverseTokens(VSPToken *tokens)
  {
    VSPToken *rev = NULL;
    for(VSPToken *tok = tokens; tok; ) {
      VSPToken *nextTok = tok->nxt;
      ADD_TO_LIST(rev, tok);
      tok = nextTok;
    }
    return rev;
  }

  static VSPToken *readTokens(VSP *sp)
  {
    FILE *cfg = NULL;
    if((cfg = fopen(sp->configFile, "r")) == NULL) {
      syslog(LOG_ERR,"cannot open config file %s : %s", sp->pidFile, strerror(errno));
      return NO;
    }

    // collect the tokens in a (reversed) list
    VSPToken *tokens = newToken("start", 5);
    char line[VSP_MAX_LINELEN];
    u_int32_t lineNo = 0;
    while(fgets(line, VSP_MAX_LINELEN, cfg)) {
      lineNo++;
      char *p = line;
      // comments start with '#'
      p[strcspn(p, "#")] = '\0';
      VSPToken *tok;
      while((tok = nextToken(p, &p)) != NULL) {
	tok->lineNo = lineNo;
	ADD_TO_LIST(tokens, tok);
      }
    }
    fclose(cfg);

    // get them in the right order
    tokens = reverseTokens(tokens);

    return tokens;
  }

  /*_________________---------------------------__________________
    _________________      readConfigFile       __________________
    -----------------___________________________------------------
  */

  int VSPReadConfigFile(VSP *sp)
  {
    EnumVSPObject level[VSP_MAX_CONFIG_DEPTH + 5];
    int depth = 0;
    level[depth] = VSPOBJ_VSP;

    // could have used something like bison to make a complete parser with
    // strict rules,  but for simplicity we just allow the current object
    // to double as a state variable that determines what is allowed next.
    
    for(VSPToken *tok = readTokens(sp); tok; tok = tok->nxt) {

      if(depth > VSP_MAX_CONFIG_DEPTH) {
	// depth overrun
	parseError(sp, tok, "too many '{'s", "");
	return NO;
      }
      else if(tok->stok == VSPTOKEN_ENDOBJ) {
	// end of level, pop the stack
	if(depth > 0) --depth;
	else {
	  parseError(sp, tok, "too many '}'s ", "");
	  return NO;
	}
      }
      else switch(level[depth]) {
      case VSPOBJ_VSP:
	// must start by opening an sFlow object
	if((tok = expectToken(sp, tok, VSPTOKEN_SFLOW)) == NULL) return NO;
	if((tok = expectToken(sp, tok, VSPTOKEN_STARTOBJ)) == NULL) return NO;
	newSFlow(sp);
	level[++depth] = VSPOBJ_SFLOW;
	break;

      case VSPOBJ_SFLOW:

	switch(tok->stok) {
	case VSPTOKEN_VIRTUALSWITCH:
	  if((tok = expectToken(sp, tok, VSPTOKEN_STARTOBJ)) == NULL) return NO;
	  newSwitch(sp->sFlow);
	  level[++depth] = VSPOBJ_SWITCH;
	  break;
	case VSPTOKEN_COLLECTOR:
	  if((tok = expectToken(sp, tok, VSPTOKEN_STARTOBJ)) == NULL) return NO;
	  newCollector(sp->sFlow);
	  level[++depth] = VSPOBJ_COLLECTOR;
	  break;
	case VSPTOKEN_FIRSTSUBAGENTID:
	  if((tok = expectInteger32(sp, tok, &sp->sFlow->firstSubAgentId, 0, 65535)) == NULL) return NO;
	  break;
	case VSPTOKEN_FIRSTIFINDEX:
	  if((tok = expectInteger32(sp, tok, &sp->sFlow->firstIfIndex, 0, 65535)) == NULL) return NO;
	  break;
	case VSPTOKEN_PACKETSAMPLINGRATE:
	  if((tok = expectInteger32(sp, tok, &sp->sFlow->sFlowSettings->samplingRate, 1, 65535)) == NULL) return NO;
	  break;
	case VSPTOKEN_COUNTERPOLLINGINTERVAL:
	  if((tok = expectInteger32(sp, tok, &sp->sFlow->sFlowSettings->pollingInterval, 1, 300)) == NULL) return NO;
	  break;
	case VSPTOKEN_AGENTIP:
	  if((tok = expectIP(sp, tok, &sp->sFlow->agentIP, NULL)) == NULL) return NO;
	  break;
	default:
	  parseError(sp, tok, "unexpected sFlow setting", "");
	  return NO;
	  break;
	}
	break;
	
      case VSPOBJ_COLLECTOR:
	{
	  VSPCollector *col = sp->sFlow->collectors;
	  switch(tok->stok) {
	  case VSPTOKEN_IP:
	    if((tok = expectIP(sp, tok, &col->ipAddr, (struct sockaddr *)&col->sendSocket)) == NULL) return NO;
	    break;
	  case VSPTOKEN_UDPPORT:
	    if((tok = expectInteger32(sp, tok, &col->udpPort, 1, 65535)) == NULL) return NO;
	    break;
	  default:
	    parseError(sp, tok, "unexpected collector setting", "");
	    return NO;
	    break;
	  }
	}
	break;

      case VSPOBJ_SWITCH:
	{
	  VSPSwitch *sw = sp->sFlow->switches;
	  switch(tok->stok) {
	  case VSPTOKEN_PCAP:
	    if((tok = expectToken(sp, tok, VSPTOKEN_STARTOBJ)) == NULL) return NO;
	    newPCAP(sw);
	    level[++depth] = VSPOBJ_PCAP;
	    break;
	  case VSPTOKEN_UPLINK:
	    {
	      if((tok = expectToken(sp, tok, VSPTOKEN_STARTOBJ)) == NULL) return NO;
	      VSPInterface *intf = newInterface(sw);
	      intf->uplink = YES;
	      if(sw->uplink) {
		parseError(sp, tok, "uplink already defined for this switch", "");
		return NO;
	      }
	      else sw->uplink = intf;
	      level[++depth] = VSPOBJ_INTERFACE;
	    }
	    break;
	  case VSPTOKEN_VIRTUALINTERFACE:
	    if((tok = expectToken(sp, tok, VSPTOKEN_STARTOBJ)) == NULL) return NO;
	    newInterface(sw);
	    level[++depth] = VSPOBJ_INTERFACE;
	    break;
	  default:
	    parseError(sp, tok, "unexpected switch setting", "");
	    return NO;
	    break;
	  }
	}
	break;

      case VSPOBJ_PCAP:
	{
	  VSPPCAP *pktReader = sp->sFlow->switches->pktReader;
	  switch(tok->stok) {
	  case VSPTOKEN_DEVICE:
	    if((tok = expectStr(sp, tok, "device (e.g. eth0)", &pktReader->device)) == NULL) return NO;
	    break;
	  case VSPTOKEN_HEADERBYTES:
	    if((tok = expectInteger32(sp, tok, &pktReader->snaplen, 64, 10000)) == NULL) return NO;
	    break;
	  case VSPTOKEN_PROMISCUOUSMODE:
	    if((tok = expectInteger32(sp, tok, &pktReader->promiscuous, 0, 1)) == NULL) return NO;
	    break;
	  default:
	    parseError(sp, tok, "unexpected pcap setting", "");
	    return NO;
	    break;
	  }
	}
	break;

      case VSPOBJ_INTERFACE:
	{
	  VSPInterface *intf = sp->sFlow->switches->interfaces;
	  switch(tok->stok) {
	  case VSPTOKEN_MAC:
	    if((tok = expectMAC(sp, tok, &intf->mac)) == NULL) return NO;
	    break;
	  case VSPTOKEN_IFTYPE:
	    if((tok = expectInteger32(sp, tok, &intf->ifType, 0, 512)) == NULL) return NO;
	    break;
	  case VSPTOKEN_IFSPEED:
	    if((tok = expectInteger64(sp, tok, &intf->ifSpeed)) == NULL) return NO;
	    break;
	  default:
	    parseError(sp, tok, "unexpected interface setting", "");
	    return NO;
	    break;
	  }
	}
	break;
	
      default:
	parseError(sp, tok, "unexpected state", "");
      }
    }

    // OK we consumed all the tokens, but we still need to run some sanity checks to make sure
    // we have a usable configuration...

    int parseOK = YES;

    if(sp->sFlow == NULL) {
      syslog(LOG_ERR, "parse error in %s : sFlow not found", sp->configFile);
      parseOK = NO;
    }
    else {
      //////////////////////// sFlow /////////////////////////
      if(sp->sFlow->agentIP.type == 0) {
	syslog(LOG_ERR, "parse error in %s : agentIP not defined", sp->configFile);
	parseOK = NO;
      }
      if(sp->sFlow->numSwitches == 0) {
	syslog(LOG_ERR, "parse error in %s : no switches defined", sp->configFile);
	parseOK = NO;
      }
      if(sp->sFlow->numCollectors == 0) {
	syslog(LOG_ERR, "parse error in %s : no collectors defined", sp->configFile);
	parseOK = NO;
      }
      for(VSPCollector *coll = sp->sFlow->collectors; coll; coll = coll->nxt) {
	//////////////////////// collector /////////////////////////
	if(coll->ipAddr.type == 0) {
	  syslog(LOG_ERR, "parse error in %s : collector  has no IP", sp->configFile);
	  parseOK = NO;
	}
      }
      for(VSPSwitch *sw = sp->sFlow->switches; sw; sw = sw->nxt) {
	//////////////////////// switch /////////////////////////
	if(sw->numInterfaces == 0 && sp->debug) {
	  syslog(LOG_ERR, "parse error in %s : switch %d has no interfaces", sp->configFile, (sw->subAgentId+1));
	  parseOK = NO;
	}
	if(sw->uplink == NULL && sp->debug) {
	  syslog(LOG_ERR, "parse error in %s : switch %d has no uplink defined", sp->configFile, (sw->subAgentId+1));
	  parseOK = NO;
	}

	if(sw->pktReader == NULL) {
	  syslog(LOG_ERR, "parse error in %s : switch %d has no pcap definition", sp->configFile, (sw->subAgentId+1));
	  parseOK = NO;
	}
	else {
	  //////////////////////// pktReader /////////////////////////
	  if(sw->pktReader->device == NULL) {
	    syslog(LOG_ERR, "parse error in %s : switch %d has no device specified in pcap definition", sp->configFile, (sw->subAgentId+1));
	    parseOK = NO;
	  }
	}

	for(VSPInterface *intf = sw->interfaces; intf; intf = intf->nxt) {
	  //////////////////////// interface /////////////////////////
	  if(intf->uplink == NO && intf->mac == 0) {
	    syslog(LOG_ERR, "parse error in %s : interface %d (switch %d) has no MAC address specified", sp->configFile, intf->ifIndex, (sw->subAgentId+1));
	    parseOK = NO;
	  }
	}
      }
    }

    return parseOK;
  }

#if defined(__cplusplus)
} /* extern "C" */
#endif
