/* Copyright (c) 2009 InMon Corp. ALL RIGHTS RESERVED */
/* License: http://www.inmon.com/products/virtual-probe/license.php */


#if defined(__cplusplus)
extern "C" {
#endif

#include "vsprobed.h"

// this fn doesn't always appear in the pcap.h file
extern int pcap_get_selectable_fd(pcap_t *);

  typedef enum { VSPSTATE_END, VSPSTATE_RUN, VSPSTATE_READCONFIG } EnumVSPState;

  // global - easier for signal handler
  EnumVSPState vsp_state = VSPSTATE_READCONFIG;
  int exitStatus = EXIT_SUCCESS;


  /*_________________---------------------------__________________
    _________________     agent callbacks       __________________
    -----------------___________________________------------------
  */

  static void *agentCB_alloc(void *magic, SFLAgent *agent, size_t bytes)
  {
    return calloc(1, bytes);
  }

  static int agentCB_free(void *magic, SFLAgent *agent, void *obj)
  {
    free(obj);
    return 0;
  }

  static void agentCB_error(void *magic, SFLAgent *agent, char *msg)
  {
    syslog(LOG_ERR, "sflow agent error: %s\n", msg);
  }

  void agentCB_getCounters(void *magic, SFLPoller *poller, SFL_COUNTERS_SAMPLE_TYPE *cs)
  {
    assert(poller->magic);
    VSPInterface *intf = (VSPInterface *)poller->magic;
    assert(intf->genElem.tag == SFLCOUNTERS_GENERIC);
    SFLADD_ELEMENT(cs, &intf->genElem);
    sfl_poller_writeCountersSample(poller, cs);
  }

  /*_________________---------------------------__________________
    _________________         initAgent         __________________
    -----------------___________________________------------------
  */
  
  static int initAgent(VSP *sp)
  {
    if(sp->debug) syslog(LOG_INFO,"creating sfl agent\n");
  
    assert(sp->sFlow->agentIP.type);
  
    int subId = 0;
    time_t now = time(NULL);
    for(VSPSwitch *sw = sp->sFlow->switches; sw; sw = sw->nxt) {
      sw->agent = (SFLAgent *)calloc(1, sizeof(SFLAgent));
      sfl_agent_init(sw->agent, &sp->sFlow->agentIP, ++subId, now, now, sp, agentCB_alloc, agentCB_free, agentCB_error, NULL);
      // just one receiver - we are serious about making this lightweight for now
      VSPCollector *collector = sp->sFlow->collectors;
      SFLReceiver *receiver = sfl_agent_addReceiver(sw->agent);
      u_int32_t receiverIndex = 1;
    
      // claim the receiver slot
      sfl_receiver_set_sFlowRcvrOwner(receiver, "Virtual Switch sFlow Probe");
    
      // set the timeout to infinity
      sfl_receiver_set_sFlowRcvrTimeout(receiver, 0xFFFFFFFF);

      // set the destination address
      assert(collector->ipAddr.type);
      if(collector->ipAddr.type == SFLADDRESSTYPE_IP_V6) {
	receiver->receiver6 = collector->sendSocket;
      }
      else {
	memcpy(&receiver->receiver4, &collector->sendSocket, sizeof(receiver->receiver4));
      }
      sfl_receiver_set_sFlowRcvrAddress(receiver, &collector->ipAddr);
    
      // set the destination port
      sfl_receiver_set_sFlowRcvrPort(receiver, collector->udpPort);
    
      // get the sFlow settings (only settable at sFlow level for now)
      VSPSFlowSettings *settings = sp->sFlow->sFlowSettings;
      u_int32_t samplingRate = settings ? settings->samplingRate : SFL_DEFAULT_SAMPLING_RATE;
      u_int32_t pollingInterval = settings ? settings->pollingInterval : SFL_DEFAULT_POLLING_INTERVAL;
    
      // add a single sampler to represent the whole switch (special <ifIndex>:0 datasource)
      // The alternative is to model a physical switch more closely and instantiate a separate 
      // sampler object for each interface,  but then unicasts would have to be offered to two
      // samplers, and broadcasts/multicasts would have to be offered to all of them. Doing it
      // this way with a single <ifindex>:0 sampler is much more efficient for a virtual switch,
      // and is allowed by the sFlow standard.
      SFLDataSource_instance dsi;
      SFL_DS_SET(dsi, 0, 0, 0);  // ds_class = 0, ds_index = <ifIndex>, ds_instance = 0
      sw->sampler = sfl_agent_addSampler(sw->agent, &dsi);
      sfl_sampler_set_sFlowFsPacketSamplingRate(sw->sampler, samplingRate);
      sfl_sampler_set_sFlowFsReceiver(sw->sampler, receiverIndex);
    
      // add poller instances for each interface
      for(VSPInterface *intf = sw->interfaces; intf; intf = intf->nxt) {
	// define the data source instance
	SFLDataSource_instance dsi;
	SFL_DS_SET(dsi, 0, intf->ifIndex, receiverIndex-1);  // ds_class = 0, ds_index = <ifIndex>, ds_instance = 0,1,2...
      
	// and a poller too. Give it the interface as the magic pointer to call back with,
	// so we can easily get to the counter blocks when we need to update them or send
	// out the counter-samples.
	SFLPoller *poller = sfl_agent_addPoller(sw->agent, &dsi, intf, agentCB_getCounters);
	sfl_poller_set_sFlowCpInterval(poller, pollingInterval);
	sfl_poller_set_sFlowCpReceiver(poller, receiverIndex);
      
	// initialize the counterBlock
	intf->genElem.tag = SFLCOUNTERS_GENERIC;
	intf->genElem.counterBlock.generic.ifIndex = intf->ifIndex;
	intf->genElem.counterBlock.generic.ifType = intf->ifType;
	intf->genElem.counterBlock.generic.ifSpeed = intf->ifSpeed;
	intf->genElem.counterBlock.generic.ifDirection = 1; // means full-duplex
	intf->genElem.counterBlock.generic.ifStatus = 3; // means ifAdminStatus=up, ifOperStatus=up
	intf->genElem.counterBlock.generic.ifPromiscuousMode = sw->pktReader->promiscuous;
      
      }
    }
    return YES;
  }

  /*_________________---------------------------__________________
    _________________       readPacket          __________________
    -----------------___________________________------------------
  */

  static void readPacket(u_char *magic, const struct pcap_pkthdr *pkthdr, const u_char *buf)
  {
    VSPSwitch *sw = (VSPSwitch *)magic;

    // copy the bytes into the 64-bit integers in the same was as we did in parseMAC()
    // so that they will match correctly regardless of processor architecture
    u_int64_t dstMAC=0, srcMAC=0;
    memcpy(&dstMAC, buf, 6);
    memcpy(&srcMAC, buf+6, 6);

    // test for multicast/broadcast
    int broadcast = 0;
    int multicast = 0;
    if(buf[0] & 0x01) {
      if(dstMAC == sw->mySFlow->macBroadcast) broadcast = 1;
      else multicast = 1;
    }

    VSPInterface *inputIF = NULL;
    VSPInterface *outputIF = NULL;

    // could implement this using a mac->interface hash table,  but since
    // most servers will only have a handful of VMs it's not clear that it's
    // worth the trouble.
    for(VSPInterface *intf = sw->interfaces; intf; intf = intf->nxt) {
      if(srcMAC == intf->mac) inputIF = intf;
      else if(dstMAC == intf->mac) outputIF = intf;
      else if(intf->uplink) {
	// don't accumulate outbound broadcasts or multicasts to the uplink until
	// we are sure it came from one of the macs (real or virtual) on this server
      }
      else if(broadcast) intf->genElem.counterBlock.generic.ifOutBroadcastPkts++;
      else if(multicast) intf->genElem.counterBlock.generic.ifOutMulticastPkts++;
    }

    if(inputIF) {
      // OK, it originated locally, so accumulate uplink outbound broadcasts/multicasts
      if(multicast) sw->uplink->genElem.counterBlock.generic.ifOutMulticastPkts++;
      else if(broadcast) sw->uplink->genElem.counterBlock.generic.ifOutBroadcastPkts++;
    }

    // anything that didn't match a local mac we assume to be out in the LAN
    if(inputIF == NULL) inputIF = sw->uplink;
    if(outputIF == NULL) outputIF = sw->uplink;
    // avoid double-counting (in the degenerate case with packets that do not match
    // any virtual interfaces we'll just count everything as inbound on the uplink)
    if(inputIF == outputIF) outputIF = NULL;

    // special flags for packets to or from the "internal" interface, which
    // we will mark specially as ifIndex 0x3FFFFFFF below, in keeping with
    // the sFlow standard.  The test is for the mac associated with the
    // uplink, but what we are actually modeling here is similar to the scenario
    // where a management CPU inside a switch sends or receives packets
    // internally.
    int fromInternal = (srcMAC == sw->uplink->mac);
    int toInternal = (dstMAC == sw->uplink->mac);

    // accumulate the counters - but not to or from "internal"
    if(inputIF && !fromInternal) {
      inputIF->genElem.counterBlock.generic.ifInOctets += pkthdr->len;
      if(broadcast) inputIF->genElem.counterBlock.generic.ifInBroadcastPkts++;
      else if(multicast) inputIF->genElem.counterBlock.generic.ifInMulticastPkts++;
      else inputIF->genElem.counterBlock.generic.ifInUcastPkts++;
    }
    if(outputIF && !toInternal) {
      outputIF->genElem.counterBlock.generic.ifOutOctets += pkthdr->len;
      if(!broadcast && !multicast) outputIF->genElem.counterBlock.generic.ifOutUcastPkts++;
    }

#ifndef NDEBUG
    if(sw->mySFlow->myVSP->debug > 1) {
      syslog(LOG_INFO,"readPacket: %02x%02x%02x%02x%02x%02x -> %02x%02x%02x%02x%02x%02x (len=%u hdr=%u in=%u out=%u)\n",
	     buf[6],
	     buf[7],
	     buf[8],
	     buf[9],
	     buf[10],
	     buf[11],
	     buf[0],
	     buf[1],
	     buf[2],
	     buf[3],
	     buf[4],
	     buf[5],
	     pkthdr->len,
	     pkthdr->caplen,
	     inputIF ? inputIF->ifIndex : 0,
	     outputIF ? outputIF->ifIndex : 0);
    }
#endif

    // test to see if we want to sample this packet
    if(sfl_sampler_takeSample(sw->sampler)) {

      // Phew, that's the end of the "fast-path".  We only get here when we decide to
      // sample a packet,  so it's probably at least two orders of magnitude less speed-critical
      // that the code above.  In switches made of metal and silicon that part is usually done
      // in hardware anyway. It's only in software-switches that we have to read every packet,
      // update the counters, and do our own random sampling.

      // Having said that,  there really isn't much more to do from here.  Just package the sampled
      // header and give it to the sflow library to be encoded and sent out.  Now you can see why
      // sFlow runs at wire speed on Terrabit switches :)

      // Build a flow sample with just a SFLFLOW_HEADER element in it
      SFL_FLOW_SAMPLE_TYPE fs;
      memset(&fs, 0, sizeof(fs));
      fs.input = fromInternal ? 0x3FFFFFFF : (inputIF ? inputIF->ifIndex : 0);
      fs.output = toInternal ? 0x3FFFFFFF : ((broadcast || multicast) ? 0x80000000 : (outputIF ? outputIF->ifIndex : 0));

      SFLFlow_sample_element hdrElem;
      memset(&hdrElem, 0, sizeof(hdrElem));

      hdrElem.tag = SFLFLOW_HEADER;
      hdrElem.flowType.header.header_protocol = SFLHEADER_ETHERNET_ISO8023;
      u_int32_t FCS_bytes = 4;
      // the FCS trailing bytes should be counted in the frame_length
      // but they should also be recorded in the "stripped" field.
      // assume that libpcap is not giving us the FCS.
      u_int32_t frame_len = pkthdr->len;
      hdrElem.flowType.header.frame_length = frame_len + FCS_bytes;
      hdrElem.flowType.header.stripped = FCS_bytes;
      u_int32_t header_len = pkthdr->caplen;
      if(header_len > frame_len) header_len = frame_len;
      if(header_len > (u_int32_t)sw->pktReader->snaplen) header_len = sw->pktReader->snaplen;
      hdrElem.flowType.header.header_length = header_len;
      hdrElem.flowType.header.header_bytes = (u_int8_t *)buf;
      SFLADD_ELEMENT(&fs, &hdrElem);
      // submit the sample to be encoded and sent out
      sfl_sampler_writeFlowSample(sw->sampler, &fs);
    }
  }


  /*_________________---------------------------__________________
    _________________      openPacketReaders    __________________
    -----------------___________________________------------------
  */

  static int openPacketReaders(VSP *sp)
  {
    for(VSPSwitch *sw = sp->sFlow->switches; sw; sw = sw->nxt) {
      // put an empty string into the errbuf so we can detect a warning
      // -- as described on the man page for pcap_open_live()
      sw->pktReader->errbuf[0] = '\0';

      if(sp->debug) {
	syslog(LOG_INFO,"calling pcap_open_live(dev=%s, snaplen=%d, promisc=%d)\n",
	       sw->pktReader->device,
	       sw->pktReader->snaplen,
	       sw->pktReader->promiscuous);
      }
      
      int timeout_ms = 0; // poll
      if((sw->pktReader->pcap = pcap_open_live(sw->pktReader->device,
					       sw->pktReader->snaplen,
					       sw->pktReader->promiscuous,
					       timeout_ms,
					       sw->pktReader->errbuf)) == NULL) {
	syslog(LOG_ERR, "error in pcap_open_live(): %s\n", sw->pktReader->errbuf);
	return NO;
      }

      if(sw->pktReader->errbuf[0] != '\0') {
	// there was a warning - log it and continue
	syslog(LOG_INFO, "warning in pcap_open_live(): %s\n", sw->pktReader->errbuf);
      }
      
      // ask for a file descriptor that we can use with select()
      // For details, see the man page for pcap_get_selectable_fd()
      if((sw->pktReader->fd = pcap_get_selectable_fd(sw->pktReader->pcap)) == -1) {
	syslog(LOG_ERR, "pcap_get_selectable_fd() returned -1 : %s\n", sw->pktReader->errbuf);
	return NO;
      }
      
    }
    return YES;
  }

  /*_________________---------------------------__________________
    _________________   closePacketReaders      __________________
    -----------------___________________________------------------
  */

  static void closePacketReaders(VSP *sp)
  {
    if(sp->sFlow) {
      for(VSPSwitch *sw = sp->sFlow->switches; sw; sw = sw->nxt) {
	if(sw->pktReader->pcap) pcap_close(sw->pktReader->pcap);
      }
    }
  }

  /*_________________---------------------------__________________
    _________________       freeSFlow           __________________
    -----------------___________________________------------------
  */

  static void freeSFlow(VSPSFlow *sf)
  {
    if(sf == NULL) return;
    if(sf->sFlowSettings) free(sf->sFlowSettings);
    for(VSPSwitch *sw = sf->switches; sw; ) {
      VSPSwitch *nextSW = sw->nxt;
      if(sw->pktReader) free(sw->pktReader);
      if(sw->agent) sfl_agent_release(sw->agent);
      for(VSPInterface *intf = sw->interfaces; intf; ) {
	VSPInterface *nextIF = intf->nxt;
	free(intf);
	intf = nextIF;
      }
      free(sw);
      sw = nextSW;
    }
    for(VSPCollector *coll = sf->collectors; coll; ) {
      VSPCollector *nextColl = coll->nxt;
      free(coll);
      coll = nextColl;
    }
    free(sf);
  }

  /*_________________---------------------------__________________
    _________________     setDefaults           __________________
    -----------------___________________________------------------
  */

  static void setDefaults(VSP *sp)
  {
    sp->debug = 0;
    sp->configFile = VSP_DEFAULT_CONFIGFILE;
    sp->pidFile = VSP_DEFAULT_PIDFILE;
  }

  /*_________________---------------------------__________________
    _________________      instructions         __________________
    -----------------___________________________------------------
  */

  static void instructions(char *command)
  {
    fprintf(stderr,"Usage: %s [-d] [-v] [-p PIDFile] [-f CONFIGFile] \n", command);
    fprintf(stderr,"\n\
             -d:  debug mode - do not fork as a daemon (-dd for verbose output to syslog)\n\
             -v:  print version number and exit\n\
             -H:  host probe - include process lookup for packets to or from this host/VM (uplink MAC == my MAC)\n\
     -p PIDFile:  specify PID file (default is " VSP_DEFAULT_PIDFILE ")\n\
  -f CONFIGFile:  specify config file (default is "VSP_DEFAULT_CONFIGFILE")\n\n");
  fprintf(stderr, "=============== More Information ============================================\n");
  fprintf(stderr, "| documentation         - http://www.inmon.com/products/virtual-probe       |\n");
  fprintf(stderr, "| sFlow standard        - http://www.sflow.org                              |\n");
  fprintf(stderr, "| sFlowTrend (FREE)     - http://www.inmon.com/products/sFlowTrend.php      |\n");
  fprintf(stderr, "| Traffic Sentinel      - http://www.inmon.com/products/trafficsentinel.php |\n");
  fprintf(stderr, "=============================================================================\n");

    exit(EXIT_FAILURE);
  }

  /*_________________---------------------------__________________
    _________________   processCommandLine      __________________
    -----------------___________________________------------------
  */

  static void processCommandLine(VSP *sp, int argc, char *argv[])
  {
    int in;
    while ((in = getopt(argc, argv, "dvp:f:?h")) != -1) {
      switch(in) {
      case 'd': sp->debug++; break;
      case 'v': printf("%s version %s\n", argv[0], VSP_VERSION); exit(EXIT_SUCCESS); break;
      case 'p': sp->pidFile = optarg; break;
      case 'f': sp->configFile = optarg; break;
      case '?':
      case 'h':
      default: instructions(*argv);
      }
    }
  }

  /*_________________---------------------------__________________
    _________________     signal_handler        __________________
    -----------------___________________________------------------
  */

  static void signal_handler(int sig) {
    switch(sig) {
    case SIGTERM:
      syslog(LOG_INFO,"Received SIGTERM");
      vsp_state = VSPSTATE_END;
      break;
    case SIGINT:
      syslog(LOG_INFO,"Received SIGINT");
      vsp_state = VSPSTATE_END;
      break;
    case SIGHUP:
      syslog(LOG_INFO,"Received SIGHUP - re-reading config");
      vsp_state = VSPSTATE_READCONFIG;
      break;
    default:
      syslog(LOG_INFO,"Received signal %d", sig);
      break;
    }
  }

  /*_________________---------------------------__________________
    _________________         main              __________________
    -----------------___________________________------------------
  */

  int main(int argc, char *argv[])
  {
    VSP sp;
    memset(&sp, 0, sizeof(sp));

    // open syslog
    openlog(VSP_DAEMON_NAME, LOG_CONS, LOG_USER);
    setlogmask(LOG_UPTO(LOG_DEBUG));

    // register signal handler
    signal(SIGTERM,signal_handler);
    signal(SIGINT,signal_handler); 
    signal(SIGHUP,signal_handler); 

    // init
    setDefaults(&sp);

    // read the command line
    processCommandLine(&sp, argc, argv);
      
    // don't run if we think another one is already running
    struct stat statBuf;
    if(stat(sp.pidFile, &statBuf) == 0) {
      syslog(LOG_ERR,"Another %s is already running. If this is an error, remove %s", argv[0], sp.pidFile);
      exit(EXIT_FAILURE);
    }

    if(sp.debug == 0) {
      // fork to daemonize
      pid_t pid = fork();
      if(pid < 0) {
	syslog(LOG_ERR,"Cannot fork child");
	exit(EXIT_FAILURE);
      }
      
      if(pid > 0) {
	// in parent - write pid file and exit
	FILE *f;
	if(!(f = fopen(sp.pidFile,"w"))) {
	  syslog(LOG_ERR,"Could not open the pid file %s for writing : %s", sp.pidFile, strerror(errno));
	  exit(EXIT_FAILURE);
	}
	fprintf(f,"%lu\n",(long)pid);
	if(fclose(f) == -1) {
	  syslog(LOG_ERR,"Could not close pid file %s : %s", sp.pidFile, strerror(errno));
	  exit(EXIT_FAILURE);
	}
	
	exit(EXIT_SUCCESS);
      }
      else {
	
	// in child
	umask(0);
	
	pid_t sid = setsid();
	if(sid < 0) {
	  syslog(LOG_ERR,"setsid failed");
	  exit(EXIT_FAILURE);
	}
	
	// close all file descriptors 
	int i;
	for(i=getdtablesize(); i >= 0; --i) close(i);
	// create stdin/out/err
	i = open("/dev/null",O_RDWR); // stdin
	dup(i);                       // stdout
	dup(i);                       // stderr
      }
    }

    syslog(LOG_INFO, "started");
    
    // initialize the clock so we can detect second boundaries
    time_t clk = time(NULL);

    while(vsp_state != VSPSTATE_END) {
     
      switch(vsp_state) {
      case VSPSTATE_READCONFIG:
	{
	  if(sp.sFlow) {
	    // we have been asked to re-read
	    // the config e.g. on a SIGHUP
	    closePacketReaders(&sp);
	    freeSFlow(sp.sFlow);
	    sp.sFlow = NULL;
	  }

	  if(VSPReadConfigFile(&sp)
	     && initAgent(&sp)
	     && openPacketReaders(&sp)) {
	    vsp_state = VSPSTATE_RUN;
	  }
	  else{
	    exitStatus = EXIT_FAILURE;
	    vsp_state = VSPSTATE_END;
	  }
	}
	break;
	
      case VSPSTATE_RUN:
	{
	  // read some packets (or time out)
	  fd_set readfds;
	  FD_ZERO(&readfds);
	  // set the timeout so that if all is quiet we will
	  // still loop around and check for ticks/signals about
	  // 10 times per second
	  struct timeval timeout;
	  timeout.tv_sec = 0;
	  timeout.tv_usec = 100000;
	  int max_fd = 0;
	  for(VSPSwitch *sw = sp.sFlow->switches; sw; sw = sw->nxt) {
	    int fd = sw->pktReader->fd;
	    if(fd > 0) {
	      FD_SET(fd, &readfds);
	      if(fd > max_fd) max_fd = fd;
	    }
	  }
      
	  int nfds = select(max_fd + 1,
			    &readfds,
			    (fd_set *)NULL,
			    (fd_set *)NULL,
			    &timeout);
	  // may return prematurely if a signal was caught, in which case nfds will be
	  // -1 and errno will be set to EINTR.  If we get any other error, abort.
	  if(nfds < 0 && errno != EINTR) {
	    syslog(LOG_ERR, "select() returned %d : %s\n", nfds, strerror(errno));
	    exit(EXIT_FAILURE);
	  }
	  if(nfds > 0) {
	    // got something
	    for(VSPSwitch *sw = sp.sFlow->switches; sw; sw = sw->nxt) {
	      if(FD_ISSET(sw->pktReader->fd, &readfds)) {
		if((pcap_dispatch(sw->pktReader->pcap,
				  sw->pktReader->batch,
				  readPacket,
				  (u_char *)sw)) == -1) {
		  syslog(LOG_ERR, "error in pcap_dispatch(): %s\n", pcap_geterr(sw->pktReader->pcap));
		  exit(EXIT_FAILURE);
		}
	      }
	    }
	  }
	  
	  // check for second boundaries and generate ticks for the sFlow library
	  time_t test_clk = time(NULL);
	  while(clk < test_clk) {
	    for(VSPSwitch *sw = sp.sFlow->switches; sw; sw = sw->nxt) {
	      sfl_agent_tick(sw->agent, clk);
	    }
	    clk++;
	  }
	}
      case VSPSTATE_END:
	break;
      }
    }

    // get here if terminated by a signal or a pcap error
    closePacketReaders(&sp);
    closelog();
    syslog(LOG_INFO,"stopped");
    if(sp.debug == 0) remove(sp.pidFile);
    exit(exitStatus);
  } /* main() */


#if defined(__cplusplus)
} /* extern "C" */
#endif

